;; Generated package description from tomelr.el  -*- no-byte-compile: t -*-
(define-package "tomelr" "0.3.0" "Convert S-expressions to TOML" '((emacs "26.3") (map "3.2.1") (seq "2.23")) :commit "ede27810243a3bf0872d4d30bb02a51ba9e8166f" :authors '(("Kaushal Modi" . "kaushal.modi@gmail.com")) :maintainer '("Kaushal Modi" . "kaushal.modi@gmail.com") :keywords '("data" "tools" "toml" "serialization" "config") :url "https://github.com/kaushalmodi/tomelr/")
